# -*- coding: utf-8 -*-
"""Credential fetcher for 2nd Screen app."""
import logging
import socket
import time

_LOGGER = logging.getLogger(__name__)


class Credentials:
    """The PS4 Credentials object. Masquerades as a PS4 to get credentials."""

    standby = '620 Server Standby'
    host_id = '1234567890AB'
    host_name = 'Home-Assistant'
    UDP_IP = '0.0.0.0'
    REQ_PORT = 997
    DDP_PORT = 987
    DDP_VERSION = '00020020'
    msg = None

    """
    PS4 listens on ports 987 and 997 (Priveleged).
    Must run command on python path:
    "sudo setcap 'cap_net_bind_service=+ep' /usr/bin/python3.5"
    """

    def __init__(self):
        """Init Cred Server."""
        self.iswakeup = False
        self.response = {
            'host-id': self.host_id,
            'host-type': 'PS4',
            'host-name': self.host_name,
            'host-request-port': self.REQ_PORT
        }
        self.start()

    def start(self):
        """Start Cred Server."""
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            self.sock = sock
        except socket.error:
            _LOGGER.error("Failed to create socket")
            return
        sock.settimeout(30)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        try:
            sock.bind((self.UDP_IP, self.DDP_PORT))
        except socket.error as error:
            _LOGGER.error(
                "Could not bind to port %s; \
                Ensure port is accessible and unused, %s",
                self.DDP_PORT, error)
            return

    def listen(self, timeout=120):
        """Listen and respond to requests."""
        start_time = time.time()
        sock = self.sock
        pings = 0
        while pings < 10:
            if timeout < time.time() - start_time:
                return
            try:
                response = sock.recvfrom(1024)
            except socket.error:
                sock.close()
                pings += 1
            data = response[0]
            address = response[1]
            if not data:
                pings += 1
                break
            if parse_ddp_response(data, 'search') == 'search':
                _LOGGER.debug("Search from: %s", address)
                msg = self.get_ddp_message(self.standby, self.response)
                sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
                try:
                    sock.sendto(msg.encode('utf-8'), address)
                except socket.error:
                    sock.close()
            if parse_ddp_response(data, 'wakeup') == 'wakeup':
                self.iswakeup = True
                _LOGGER.debug("Wakeup from: %s", address)
                creds = get_creds(data)
                sock.close()
                return creds
        return

    def get_ddp_message(self, status, data=None):
        """Get DDP message."""
        msg = u'HTTP/1.1 {}\n'.format(status)
        if data:
            for key, value in data.items():
                msg += u'{}:{}\n'.format(key, value)
        msg += u'device-discovery-protocol-version:{}\n'.format(
            self.DDP_VERSION)
        return msg


def parse_ddp_response(response, listen_type):
    """Parse the response."""
    rsp = response.decode('utf-8')
    if listen_type == 'search':
        if 'SRCH' in rsp:
            return 'search'
    elif listen_type == 'wakeup':
        if 'WAKEUP' in rsp:
            return 'wakeup'


def get_creds(response):
    """Return creds."""
    keys = {}
    data = response.decode('utf-8')
    for line in data.splitlines():
        line = line.strip()
        if ":" in line:
            value = line.split(':')
            keys[value[0]] = value[1]
    cred = keys['user-credential']
    return cred



"""Exceptions Definitions."""

class NotReady(Exception):
    """PS4 no connection."""
    pass


class LoginFailed(Exception):
    """PS4 Failed Login."""
    pass


class UnknownButton(Exception):
    """Button not valid."""
    pass



# -*- coding: utf-8 -*-
"""Methods for PS4 Object."""
from __future__ import print_function
from threading import Timer
import json
import logging
import time
import socket

from .connection import Connection
from .ddp import get_status, launch, wakeup
from .errors import NotReady, UnknownButton, LoginFailed
from .media_art import get_ps_store_data as ps_data
from .media_art import (search_all, COUNTRIES, DEPRECATED_REGIONS)

_LOGGER = logging.getLogger(__name__)


def open_credential_file(filename):
    """Open credential file."""
    return json.load(open(filename))


class StatusTimer():
    """Status Thread Class."""

    def __init__(self, seconds, target):
        """Init."""
        self._should_continue = False
        self.is_running = False
        self.seconds = seconds
        self.target = target
        self.thread = None

    def _handle_target(self):
        self.is_running = True
        self.target()
        self.is_running = False
        self._start_timer()

    def _start_timer(self):
        """Restart Timer."""
        if self._should_continue:
            self.thread = Timer(self.seconds, self._handle_target)
            self.thread.start()

    def start(self):
        """Start Timer."""
        if not self._should_continue and not self.is_running:
            self._should_continue = True
            self._start_timer()

    def cancel(self):
        """Cancel Timer."""
        if self.thread is not None:
            self._should_continue = False
            self.thread.cancel()


class Ps4(object):
    """The PS4 object."""

    STATUS_OK = 200
    STATUS_STANDBY = 620

    def __init__(self, host, credential=None, credentials_file=None,
                 broadcast=False):
        """Initialize the instance.

        Keyword arguments:
            host -- the host IP address
            credential -- the credential string
            credential_file -- the credendtial file generated with ps4-waker
            broadcast -- use broadcast IP address (default False)
        """
        self._host = host
        self._broadcast = broadcast
        self._socket = None
        self._credential = None
        self._connected = False
        self._status_timer = None
        self.keep_alive = False

        if credential:
            self._credential = credential
        if credentials_file:
            creds = open_credential_file(credentials_file)
            self._credential = creds['user-credential']

        self._connection = Connection(host, credential=self._credential)

    def open(self):
        """Open a connection to the PS4."""
        if self.is_standby():
            raise NotReady

        if self._connected is False:
            self.wakeup()
            self.launch()
            time.sleep(0.5)
            self._connection.connect()
            login = self._connection.login()
            if login is True:
                self._connected = True
                if self.keep_alive is True:
                    _LOGGER.debug("Keep Alive feature enabled")
                    self._status_timer = StatusTimer(30, self.send_status)
                    self._status_timer.start()

    def close(self):
        """Close the connection to the PS4."""
        self._connection.disconnect()
        self._connected = False
        self.keep_alive = False
        if self._status_timer is not None:
            self._status_timer.cancel()
            self._status_timer = None

    def is_keepalive(self):
        """Check if keep alive should be sent."""
        if self.keep_alive is False:
            self.close()

    def get_status(self):
        """Get current status info.

        Return a dictionary with status information.
        """
        return get_status(self._host)

    def launch(self):
        """Launch."""
        launch(self._host, self._credential)

    def wakeup(self):
        """Wakeup."""
        wakeup(self._host, self._credential)

    def login(self, pin=None):
        """Login."""
        self.open()
        is_login = self._connection.login(pin)
        if is_login is False:
            raise LoginFailed
        self.close()
        return is_login

    def standby(self):
        """Standby."""
        if self._connected is True:
            self.close()
        self.open()
        self._connection.standby()
        self.close()

    def start_title(self, title_id, running_id=None):
        """Start title.

        `title_id`: title to start
        """
        self.open()
        self._connection.start_title(title_id)
        if running_id is not None:
            self._connection.remote_control(16, 0)
        elif running_id == title_id:
            _LOGGER.warning("Title: %s already started", title_id)
        self.is_keepalive()

    def remote_control(self, button_name, hold_time=0):
        """Send a remote control button press.

        Documentation from ps4-waker source:
        near as I can tell, here's how this works:
         - For a simple tap, you send the key with holdTime=0,
           followed by KEY_OFF and holdTime = 0
         - For a long press/hold, you still send the key with
           holdTime=0, the follow it with the key again, but
           specifying holdTime as the hold duration.
         - After sending a direction, you should send KEY_OFF
           to clean it up (since it can just be held forever).
           Doing this after a long-press of PS just breaks it,
           however.
        """
        buttons = {'up': 1,
                   'down': 2,
                   'right': 4,
                   'left': 8,
                   'enter': 16,
                   'back': 32,
                   'option': 64,
                   'ps': 128,
                   'key_off': 256,
                   'cancel': 512,
                   'open_rc': 1024,
                   'close_rc': 2048}
        button_name = button_name.lower()
        if button_name not in buttons.keys():
            raise UnknownButton
        else:
            operation = buttons[button_name]
        self.open()
        self._connection.remote_control(operation, hold_time)
        self.is_keepalive()

    def send_status(self):
        """Send connection status to PS4."""
        if self._connected is True:
            is_loggedin = self._connection.send_status()
            if is_loggedin is False:
                self.close()

    def get_host_status(self):
        """Get PS4 status code.

        STATUS_OK: 200
        STATUS_STANDBY: 620
        """
        return self.get_status()['status_code']

    def is_running(self):
        """Return if the PS4 is running.

        Returns True or False.
        """
        return True if self.get_host_status() == self.STATUS_OK else False

    def is_standby(self):
        """Return if the PS4 is in standby.

        Returns True or False.
        """
        return True if self.get_host_status() == self.STATUS_STANDBY else False

    def get_system_version(self):
        """Get the system version."""
        return self.get_status()['system-version']

    def get_host_id(self):
        """Get the host id."""
        return self.get_status()['host-id']

    def get_host_name(self):
        """Get the host name."""
        return self.get_status()['host-name']

    def get_running_app_titleid(self):
        """Return the title Id of the running application."""
        return self.get_status()['running-app-titleid']

    def get_running_app_name(self):
        """Return the name of the running application."""
        return self.get_status()['running-app-name']

    def get_ps_store_data(self, title, title_id, region, url=None):
        """Return Title and Cover data."""
        regions = COUNTRIES
        d_regions = DEPRECATED_REGIONS

        if region not in regions:
            if region in d_regions:
                _LOGGER.warning('Region: %s is deprecated', region)
                region = d_regions[region]
            else:
                _LOGGER.error('Region: %s is not valid', region)
                return
        else:
            region = regions[region]
        try:
            _title, art = ps_data(self, title, title_id, region, url=None)
        except TypeError:
            _LOGGER.debug("Could not find title in default database.")
            try:
                _title, art = search_all(title, title_id)
            except TypeError:
                _LOGGER.warning("Could not find cover art for: %s", title)
                return None, None
        finally:
            _LOGGER.debug("Found Title: %s, URL: %s", _title, art)
            return _title, art



# -*- coding: utf-8 -*-
from __future__ import print_function

import re
import socket
import logging

_LOGGER = logging.getLogger(__name__)

UDP_IP = '0.0.0.0'
UDP_PORT = 0

DDP_PORT = 987
DDP_VERSION = '00020020'


def get_ddp_message(msg_type, data=None):
    """Get DDP message."""
    msg = u'{} * HTTP/1.1\n'.format(msg_type)
    if data:
        for key, value in data.items():
            msg += '{}:{}\n'.format(key, value)
    msg += 'device-discovery-protocol-version:{}\n'.format(DDP_VERSION)
    return msg


def parse_ddp_response(rsp):
    """Parse the response."""
    data = {}
    app_name = None
    for line in rsp.splitlines():
        if 'running-app-name' in line:
            app_name = line
            app_name = app_name.replace('running-app-name:', '')
        re_status = re.compile(r'HTTP/1.1 (?P<code>\d+) (?P<status>.*)')
        line = line.strip()
        # skip empty lines
        if not line:
            continue
        elif re_status.match(line):
            data[u'status_code'] = int(re_status.match(line).group('code'))
            data[u'status'] = re_status.match(line).group('status')
        else:
            values = line.split(':')
            data[values[0]] = values[1]
    if app_name is not None:
        data['running-app-name'] = app_name
    return data


def get_ddp_search_message():
    """Get DDP search message."""
    return get_ddp_message('SRCH')


def get_ddp_wake_message(credential):
    """Get DDP wake message."""
    data = {
        'user-credential': credential,
        'client-type': 'a',
        'auth-type': 'C',
    }
    return get_ddp_message('WAKEUP', data)


def get_ddp_launch_message(credential):
    """Get DDP launch message."""
    data = {
        'user-credential': credential,
        'client-type': 'a',
        'auth-type': 'C',
    }
    return get_ddp_message('LAUNCH', data)


def _send_recv_msg(host, broadcast, msg, receive=True):
    """Send a ddp message and receive the response."""
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    sock.bind((UDP_IP, UDP_PORT))
    sock.settimeout(3.0)

    if broadcast:
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        _host = host or '255.255.255.255'
    else:
        _host = host

    sock.sendto(msg.encode('utf-8'), (_host, DDP_PORT))

    if receive:
        return sock.recvfrom(1024)


def _send_msg(host, broadcast, msg):
    """Send a ddp message."""
    _send_recv_msg(host, broadcast, msg, receive=False)


def search(host=None, broadcast=True):
    """Discover PS4s."""
    msg = get_ddp_search_message()
    data, addr = _send_recv_msg(host, broadcast, msg)
    if data is not None:
        ps_list = []
        data = parse_ddp_response(data.decode('utf-8'))
        data[u'host-ip'] = addr[0]
        ps_list.append(data)
        return ps_list


def get_status(host):
    """Get status."""
    ps_list = search(host=host)
    return ps_list[0]


def wakeup(host, credential, broadcast=None):
    """Wakeup PS4s."""
    msg = get_ddp_wake_message(credential)
    _send_msg(host, broadcast, msg)


def launch(host, credential, broadcast=None):
    """Launch."""
    msg = get_ddp_launch_message(credential)
    _send_msg(host, broadcast, msg)


class Discovery:
    """Device Discovery server."""

    def __init__(self):
        """Init."""
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sock.settimeout(6.0)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        self.msg = get_ddp_search_message()
        self.host = '255.255.255.255'
        self.ps_list = []

    def search(self, host):
        """Search for Devices."""
        if host is None:
            host = self.host
        finished = False
        try:
            self.send(host)
        except (socket.error, socket.timeout):
            self.sock.close
            return
        while finished is False:
            try:
                device = self.receive()
                if device not in self.ps_list:
                    self.ps_list.append(device)
            except (socket.error, socket.timeout):
                return self.ps_list

    def send(self, host):
        """Broadcast Message."""
        self.sock.sendto(self.msg.encode('utf-8'), (host, DDP_PORT))

    def receive(self):
        """Receive Message."""
        data, addr = self.sock.recvfrom(1024)
        if data is not None:
            data = parse_ddp_response(data.decode('utf-8'))
            data[u'host-ip'] = addr[0]
            return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function

import argparse
import logging
import pprint
import sys

import pyps4

try:
    from version import __version__
except ImportError:
    __version__ = 'dev'


_LOGGER = logging.getLogger(__name__)


def cmd_search(_, __):
    """The search command."""
    info = pyps4.search(broadcast=True)
    pprint.pprint(info)


def cmd_status(playstation, _):
    """The status command."""
    info = playstation.get_status()
    pprint.pprint(info)


def cmd_launch(playstation, _):
    """The status command."""
    playstation.launch()


def cmd_wakeup(playstation, _):
    """Wakeup the PS4."""
    playstation.wakeup()


def cmd_login(playstation, _):
    """Login the PS4."""
    try:
        playstation.login()
    except pyps4.NotReady:
        print('playstaion not ready')
        sys.exit(1)


def cmd_standby(playstation, _):
    """Set the PS4 in standby."""
    try:
        playstation.standby()
    except pyps4.NotReady:
        print('playstaion not ready')
        sys.exit(1)


def cmd_start_title(playstation, args):
    """Set the PS4 in standby."""
    try:
        playstation.start_title(args.title_id)
    except pyps4.NotReady:
        print('playstaion not ready')
        sys.exit(1)


def cmd_remote_control(playstation, args):
    """Send a remote control button."""
    try:
        playstation.remote_control(args.button, args.hold_time)
    except pyps4.NotReady:
        print('playstaion not ready')
        sys.exit(1)


def main(args=None):
    """The main function."""
    parser = argparse.ArgumentParser(
        description='PS4 CLI tool.')
    parser.add_argument('-v', action='store_true', dest='verbose',
                        help='be more verbose')
    parser.add_argument('-H', '--host', type=str, dest='host',
                        help='PS4 IP address', default=None)
    parser.add_argument('-c', '--credential_file', type=str,
                        dest='credential_file', default=None,
                        help='The credential file')
    parser.add_argument('-C', '--credential', type=str,
                        dest='credential', default=None,
                        help='The credential as string')
    parser.add_argument('-V', '--version', action='version',
                        version='{version}'.format(version=__version__),
                        help='Print version')

    _sub = parser.add_subparsers(title='Commands')
    _sub.required = True

    # search all devices
    subparser = _sub.add_parser('search', help='Search for PS4 devices')
    subparser.set_defaults(func=cmd_search)

    # info
    subparser = _sub.add_parser('status', help='Show current status')
    subparser.set_defaults(func=cmd_status)

    # launch
    subparser = _sub.add_parser('launch', help='Show current status')
    subparser.set_defaults(func=cmd_launch)

    # wake
    subparser = _sub.add_parser('wakeup', help='Wakeup the PS4')
    subparser.set_defaults(func=cmd_wakeup)

    # login
    subparser = _sub.add_parser('login', help='Login the PS4')
    subparser.set_defaults(func=cmd_login)

    # standby
    subparser = _sub.add_parser('standby', help='Standby the PS4')
    subparser.set_defaults(func=cmd_standby)

    # start
    subparser = _sub.add_parser('start', help='Start a title')
    subparser.add_argument('title_id', type=str,
                           metavar="TITLE", help='Game title')
    subparser.set_defaults(func=cmd_start_title)

    # remote
    subparser = _sub.add_parser('remote', help='Send remote control')
    subparser.add_argument('button', type=str,
                           metavar="BUTTON", help='button')
    subparser.add_argument('hold_time', type=int, default=0,
                           metavar="HOLD_TIME", help='hold time')
    subparser.set_defaults(func=cmd_remote_control)

    args = parser.parse_args(args)

    playstation = None

    if not hasattr(args, 'func'):
        parser.print_help()
        sys.exit()

    if args.verbose:
        logging.basicConfig()
        logging.getLogger('pyps4').setLevel(logging.DEBUG)

    try:
        playstation = pyps4.Ps4(args.host, credential=args.credential,
                                credentials_file=args.credential_file)
        args.func(playstation, args)
    finally:
        pass


if __name__ == '__main__':
    main()



"""Helpers."""
import logging

import pyps4_homeassistant as pyps4

_LOGGER = logging.getLogger(__name__)


class Helper:
    """Helpers for PS4."""

    def __init__(self):
        """Init Class."""
        pass

    def has_devices(self, host=None):
        """Return if there are devices that can be discovered."""
        from pyps4_homeassistant import ddp

        _LOGGER.debug("Searching for PS4 Devices")
        discover = ddp.Discovery()
        devices = discover.search(host)
        for device in devices:
            _LOGGER.debug("Found PS4 at: %s", device['host-ip'])
        return devices

    def link(self, host, creds, pin):
        """Perform pairing with PS4."""
        ps4 = pyps4.Ps4(host, creds)
        is_ready = True
        is_login = True
        try:
            ps4.login(pin)
        except pyps4.errors.NotReady:
            is_ready = False
        except pyps4.errors.LoginFailed:
            is_login = False
        return is_ready, is_login

    def get_creds(self):
        """Return Credentials."""
        credentials = pyps4.Credentials()
        return credentials.listen()

    def port_bind(self, ports):
        """Try binding to ports."""
        import socket
        for port in ports:
            try:
                sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                sock.settimeout(3.0)
                sock.bind(('0.0.0.0', port))
                sock.close()
            except socket.error:
                sock.close()
                return int(port)
        return



# -*- coding: utf-8 -*-
from __future__ import print_function
from threading import Timer
import json
import logging
import time
import socket

from .connection import Connection
from .ddp import get_status, launch, wakeup
from .errors import NotReady, UnknownButton, LoginFailed
from .media_art import get_ps_store_data as ps_data
from .media_art import search_all

_LOGGER = logging.getLogger(__name__)


def open_credential_file(filename):
    """Open credential file."""
    return json.load(open(filename))


class StatusTimer():
    """Status Thread Class."""

    def __init__(self, seconds, target):
        """Init."""
        self._should_continue = False
        self.is_running = False
        self.seconds = seconds
        self.target = target
        self.thread = None

    def _handle_target(self):
        self.is_running = True
        self.target()
        self.is_running = False
        self._start_timer()

    def _start_timer(self):
        """Restart Timer."""
        if self._should_continue:
            self.thread = Timer(self.seconds, self._handle_target)
            self.thread.start()

    def start(self):
        """Start Timer."""
        if not self._should_continue and not self.is_running:
            self._should_continue = True
            self._start_timer()

    def cancel(self):
        """Cancel Timer."""
        if self.thread is not None:
            self._should_continue = False
            self.thread.cancel()


class Ps4(object):
    """The PS4 object."""

    STATUS_OK = 200
    STATUS_STANDBY = 620

    def __init__(self, host, credential=None, credentials_file=None,
                 broadcast=False):
        """Initialize the instance.

        Keyword arguments:
            host -- the host IP address
            credential -- the credential string
            credential_file -- the credendtial file generated with ps4-waker
            broadcast -- use broadcast IP address (default False)
        """
        self._host = host
        self._broadcast = broadcast
        self._socket = None
        self._credential = None
        self._connected = False
        self._status_timer = None
        self.keep_alive = False

        if credential:
            self._credential = credential
        if credentials_file:
            creds = open_credential_file(credentials_file)
            self._credential = creds['user-credential']

        self._connection = Connection(host, credential=self._credential)

    def open(self):
        """Open a connection to the PS4."""
        if self.is_standby():
            raise NotReady

        if self._connected is False:
            self.wakeup()
            self.launch()
            time.sleep(0.5)
            self._connection.connect()
            login = self._connection.login()
            if login is True:
                self._connected = True
                if self.keep_alive is True:
                    _LOGGER.debug("Keep Alive feature enabled")
                    self._status_timer = StatusTimer(30, self.send_status)
                    self._status_timer.start()

    def close(self):
        """Close the connection to the PS4."""
        self._connection.disconnect()
        self._connected = False
        self.keep_alive = False
        if self._status_timer is not None:
            self._status_timer.cancel()
            self._status_timer = None

    def is_keepalive(self):
        """Check if keep alive should be sent."""
        if self.keep_alive is False:
            self.close()

    def get_status(self):
        """Get current status info.

        Return a dictionary with status information.
        """
        return get_status(self._host)

    def launch(self):
        """Launch."""
        launch(self._host, self._credential)

    def wakeup(self):
        """Wakeup."""
        wakeup(self._host, self._credential)

    def login(self, pin=None):
        """Login."""
        self.open()
        is_login = self._connection.login(pin)
        if is_login is False:
            raise LoginFailed
        self.close()
        return is_login

    def standby(self):
        """Standby."""
        if self._connected is True:
            self.close()
        self.open()
        self._connection.standby()
        self.close()

    def start_title(self, title_id, running_id=None):
        """Start title.

        `title_id`: title to start
        """
        self.open()
        self._connection.start_title(title_id)
        if running_id is not None:
            self._connection.remote_control(16, 0)
        elif running_id == title_id:
            _LOGGER.warning("Title: %s already started", title_id)
        self.is_keepalive()

    def remote_control(self, button_name, hold_time=0):
        """Send a remote control button press.

        Documentation from ps4-waker source:
        near as I can tell, here's how this works:
         - For a simple tap, you send the key with holdTime=0,
           followed by KEY_OFF and holdTime = 0
         - For a long press/hold, you still send the key with
           holdTime=0, the follow it with the key again, but
           specifying holdTime as the hold duration.
         - After sending a direction, you should send KEY_OFF
           to clean it up (since it can just be held forever).
           Doing this after a long-press of PS just breaks it,
           however.
        """
        buttons = []
        buttons.append(button_name)

        self.open()

        for button in buttons:
            try:
                operation = {
                    'up': 1,
                    'down': 2,
                    'right': 4,
                    'left': 8,
                    'enter': 16,
                    'back': 32,
                    'option': 64,
                    'ps': 128,
                    'key_off': 256,
                    'cancel': 512,
                    'open_rc': 1024,
                    'close_rc': 2048,
                }[button.lower()]
            except KeyError:
                raise UnknownButton

            self._connection.remote_control(operation, hold_time)
        self.is_keepalive()

    def send_status(self):
        """Send connection status to PS4."""
        if self._connected is True:
            is_loggedin = self._connection.send_status()
            if is_loggedin is False:
                self.close()

    def get_host_status(self):
        """Get PS4 status code.

        STATUS_OK: 200
        STATUS_STANDBY: 620
        """
        return self.get_status()['status_code']

    def is_running(self):
        """Return if the PS4 is running.

        Returns True or False.
        """
        return True if self.get_host_status() == self.STATUS_OK else False

    def is_standby(self):
        """Return if the PS4 is in standby.

        Returns True or False.
        """
        return True if self.get_host_status() == self.STATUS_STANDBY else False

    def get_system_version(self):
        """Get the system version."""
        return self.get_status()['system-version']

    def get_host_id(self):
        """Get the host id."""
        return self.get_status()['host-id']

    def get_host_name(self):
        """Get the host name."""
        return self.get_status()['host-name']

    def get_running_app_titleid(self):
        """Return the title Id of the running application."""
        return self.get_status()['running-app-titleid']

    def get_running_app_name(self):
        """Return the name of the running application."""
        return self.get_status()['running-app-name']

    def get_ps_store_data(self, title, title_id, region, url=None):
        """Return Title and Cover data."""
        regions = {'R1': 'en/US', 'R2': 'en/GB',
                   'R3': 'en/HK', 'R4': 'en/AU', 'R5': 'en/IN'}

        if region not in regions:
            _LOGGER.error('Region: %s is not valid', region)
            return
        else:
            region = regions[region]
        try:
            _title, art = ps_data(self, title, title_id, region, url=None)
        except TypeError:
            _LOGGER.debug("Could not find title in default database.")
            _title, art = search_all(title, title_id)
        finally:
            return _title, art


class Credentials:
    """The PS4 Credentials object. Masquerades as a PS4 to get credentials."""

    standby = '620 Server Standby'
    host_id = '1234567890AB'
    host_name = 'Home-Assistant'
    UDP_IP = '0.0.0.0'
    REQ_PORT = 997
    DDP_PORT = 987
    DDP_VERSION = '00020020'
    msg = None

    """
    PS4 listens on ports 987 and 997 (Priveleged).
    Must run command on python path:
    "sudo setcap 'cap_net_bind_service=+ep' /usr/bin/python3.5"
    """

    def __init__(self):
        """Init Cred Server."""
        self.iswakeup = False
        self.response = {
            'host-id': self.host_id,
            'host-type': 'PS4',
            'host-name': self.host_name,
            'host-request-port': self.REQ_PORT
        }
        self.start()

    def start(self):
        """Start Cred Server."""
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            self.sock = sock
        except socket.error:
            _LOGGER.error("Failed to create socket")
            return
        sock.settimeout(30)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        try:
            sock.bind((self.UDP_IP, self.DDP_PORT))
        except socket.error as error:
            _LOGGER.error(
                "Could not bind to port %s; \
                Ensure port is accessible and unused, %s",
                self.DDP_PORT, error)
            return

    def listen(self, timeout=120):
        """Listen and respond to requests."""
        start_time = time.time()
        sock = self.sock
        pings = 0
        while pings < 10:
            if timeout < time.time() - start_time:
                return
            try:
                response = sock.recvfrom(1024)
            except socket.error:
                sock.close()
                pings += 1
            data = response[0]
            address = response[1]
            if not data:
                pings += 1
                break
            if parse_ddp_response(data, 'search') == 'search':
                _LOGGER.debug("Search from: %s", address)
                msg = self.get_ddp_message(self.standby, self.response)
                sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
                try:
                    sock.sendto(msg.encode('utf-8'), address)
                except socket.error:
                    sock.close()
            if parse_ddp_response(data, 'wakeup') == 'wakeup':
                self.iswakeup = True
                _LOGGER.debug("Wakeup from: %s", address)
                creds = get_creds(data)
                sock.close()
                return creds
        return

    def get_ddp_message(self, status, data=None):
        """Get DDP message."""
        msg = u'HTTP/1.1 {}\n'.format(status)
        if data:
            for key, value in data.items():
                msg += u'{}:{}\n'.format(key, value)
        msg += u'device-discovery-protocol-version:{}\n'.format(
            self.DDP_VERSION)
        return msg


def parse_ddp_response(response, listen_type):
    """Parse the response."""
    rsp = response.decode('utf-8')
    if listen_type == 'search':
        if 'SRCH' in rsp:
            return 'search'
    elif listen_type == 'wakeup':
        if 'WAKEUP' in rsp:
            return 'wakeup'


def get_creds(response):
    """Return creds."""
    keys = {}
    data = response.decode('utf-8')
    for line in data.splitlines():
        line = line.strip()
        if ":" in line:
            value = line.split(':')
            keys[value[0]] = value[1]
    cred = keys['user-credential']
    return cred



# -*- coding: utf-8 -*-
from .connection import *
from .ddp import *
from .errors import *
from .ps4 import *
from .credential import *
from .helpers import *
from .media_art import *
from .const import *



# -*- coding: utf-8 -*-
from __future__ import print_function

import binascii
import logging
import socket
import time

from construct import (Bytes, Const, Int32ul, Padding, Struct)
from Cryptodome.Cipher import AES, PKCS1_OAEP
from Cryptodome.PublicKey import RSA

_LOGGER = logging.getLogger(__name__)

PUBLIC_KEY = (
    '-----BEGIN PUBLIC KEY-----\n'
    'MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAxfAO/MDk5ovZpp7xlG9J\n'
    'JKc4Sg4ztAz+BbOt6Gbhub02tF9bryklpTIyzM0v817pwQ3TCoigpxEcWdTykhDL\n'
    'cGhAbcp6E7Xh8aHEsqgtQ/c+wY1zIl3fU//uddlB1XuipXthDv6emXsyyU/tJWqc\n'
    'zy9HCJncLJeYo7MJvf2TE9nnlVm1x4flmD0k1zrvb3MONqoZbKb/TQVuVhBv7SM+\n'
    'U5PSi3diXIx1Nnj4vQ8clRNUJ5X1tT9XfVmKQS1J513XNZ0uYHYRDzQYujpLWucu\n'
    'ob7v50wCpUm3iKP1fYCixMP6xFm0jPYz1YQaMV35VkYwc40qgk3av0PDS+1G0dCm\n'
    'swIDAQAB\n'
    '-----END PUBLIC KEY-----')


def _get_public_key_rsa():
    key = RSA.importKey(PUBLIC_KEY)
    return key.publickey()


class Connection(object):
    """The TCP connection class."""

    def __init__(self, host, credential=None, port=997):
        """Init Class."""
        self._host = host
        self._credential = credential
        self._port = port
        self._socket = None
        self._cipher = None
        self._decipher = None
        self._random_seed = None

    def connect(self):
        """Open the connection."""
        _LOGGER.debug('Connect')
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._socket.connect((self._host, self._port))
        self._random_seed = \
            b'\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        self._send_hello_request()
        data = self._recv_hello_request()
        self._set_crypto_init_vector(data.seed)
        self._send_handshake_request(data.seed)

    def disconnect(self):
        """Close the connection."""
        self._socket.close()
        self._reset_crypto_init_vector()
        self._random_seed = None

    def login(self, pin=None):
        """Login."""
        _LOGGER.debug('Login')
        if pin is None:
            self._send_login_request()
        else:
            self._send_login_request(pin)
        msg = self._recv_msg()
        msg = self._decipher.decrypt(msg)
        _LOGGER.debug('RX: %s %s', len(msg), binascii.hexlify(msg))
        return self._handle_response('login', msg)

    def standby(self):
        """Request standby."""
        _LOGGER.debug('Request standby')
        self._send_standby_request()
        msg = self._recv_msg()
        msg = self._decipher.decrypt(msg)
        _LOGGER.debug('RX: %s %s', len(msg), binascii.hexlify(msg))
        return self._handle_response('standby', msg)

    def start_title(self, title_id):
        """Start an application/game title."""
        _LOGGER.debug('Start title: %s', title_id)
        self._send_boot_request(title_id)
        msg = self._recv_msg()
        msg = self._decipher.decrypt(msg)
        _LOGGER.debug('RX: %s %s', len(msg), binascii.hexlify(msg))
        return self._handle_response('start_title', msg)

    def remote_control(self, op, hold_time=0):
        """Send remote control command."""
        _LOGGER.debug('Remote control: %s (%s)', op, hold_time)
        self._send_remote_control_request(op, hold_time)

    def send_status(self):
        """Send client connection status."""
        _LOGGER.debug('Sending Status: Connected')
        self._send_status()
        msg = self._recv_msg()
        msg = self._decipher.decrypt(msg)
        _LOGGER.debug('RX: %s %s', len(msg), binascii.hexlify(msg))
        return self._handle_response('send_status', msg)

    def _handle_response(self, command, msg):
        """Return Pass/Fail for sent message."""
        pass_response = {
            'send_status': 18,
            'remote_control': 18,
            'start_title': 11,
            'standby': 27,
            'login': [0, 17]
        }

        if command == 'login':
            response_byte = msg[8]
            if response_byte in pass_response['login']:
                _LOGGER.debug("Login Successful")
                return True
            else:
                _LOGGER.debug("Login Failed")
                return False
        else:
            response_byte = msg[4]

        _LOGGER.debug("RECV: %s for Command: %s", response_byte, command)
        if pass_response[command] != response_byte:
            _LOGGER.warning("Command: %s Failed", command)
            return False
        return True

    def _send_msg(self, msg, encrypted=False):
        _LOGGER.debug('TX: %s %s', len(msg), binascii.hexlify(msg))
        if encrypted:
            msg = self._cipher.encrypt(msg)
        self._socket.send(msg)

    def _recv_msg(self, encrypted=False):
        msg = self._socket.recv(1024)
        return msg

    def _set_crypto_init_vector(self, init_vector):
        self._cipher = AES.new(self._random_seed, AES.MODE_CBC, init_vector)
        self._decipher = AES.new(self._random_seed, AES.MODE_CBC, init_vector)

    def _reset_crypto_init_vector(self):
        self._cipher = None
        self._decipher = None

    def _send_hello_request(self):
        fmt = Struct(
            'length' / Const(b'\x1c\x00\x00\x00'),
            'type' / Const(b'\x70\x63\x63\x6f'),
            'version' / Const(b'\x00\x00\x02\x00'),
            'dummy' / Padding(16),
        )

        msg = fmt.build({})
        self._send_msg(msg)

    def _recv_hello_request(self):
        fmt = Struct(
            'length' / Int32ul,
            'type' / Int32ul,
            'version' / Int32ul,
            'dummy' / Bytes(8),
            'seed' / Bytes(16),
        )

        msg = self._recv_msg()
        data = fmt.parse(msg)
        return data

    def _send_handshake_request(self, seed):
        fmt = Struct(
            'length' / Const(b'\x18\x01\x00\x00'),
            'type' / Const(b'\x20\x00\x00\x00'),
            'key' / Bytes(256),
            'seed' / Bytes(16),
        )

        recipient_key = _get_public_key_rsa()
        cipher_rsa = PKCS1_OAEP.new(recipient_key)
        key = cipher_rsa.encrypt(self._random_seed)

        _LOGGER.debug('key %s', binascii.hexlify(key))

        msg = fmt.build({'key': key, 'seed': seed})
        self._send_msg(msg)

    def _send_login_request(self, pin=None):
        fmt = Struct(
            'length' / Const(b'\x80\x01\x00\x00'),
            'type' / Const(b'\x1e\x00\x00\x00'),
            'pass_code' / Const(b'\x00\x00\x00\x00'),
            'magic_number' / Const(b'\x01\x02\x00\x00'),
            'account_id' / Bytes(64),
            'app_label' / Bytes(256),
            'os_version' / Bytes(16),
            'model' / Bytes(16),
            'pin_code' / Bytes(16),
        )

        if pin is None:
            self.pin = b''
        else:
            self.pin = pin.encode()

        config = {
            'app_label': b'PlayStation'.ljust(256, b'\x00'),
            'account_id': self._credential.encode().ljust(64, b'\x00'),
            'os_version': b'4.4'.ljust(16, b'\x00'),
            'model': b'Home-Assistant'.ljust(16, b'\x00'),
            'pin_code': self.pin.ljust(16, b'\x00'),
        }

        _LOGGER.debug('config %s', config)

        msg = fmt.build(config)
        self._send_msg(msg, encrypted=True)

    def _send_standby_request(self):
        fmt = Struct(
            'length' / Const(b'\x08\x00\x00\x00'),
            'type' / Const(b'\x1a\x00\x00\x00'),
            'dummy' / Padding(8),
        )

        msg = fmt.build({})
        self._send_msg(msg, encrypted=True)

    def _send_boot_request(self, title_id):
        fmt = Struct(
            'length' / Const(b'\x18\x00\x00\x00'),
            'type' / Const(b'\x0a\x00\x00\x00'),
            'title_id' / Bytes(16),
            'dummy' / Padding(8),
        )

        msg = fmt.build({'title_id': title_id.encode().ljust(16, b'\x00')})
        self._send_msg(msg, encrypted=True)

    def _send_remote_control_request(self, op, hold_time=0):
        fmt = Struct(
            'length' / Const(b'\x10\x00\x00\x00'),
            'type' / Const(b'\x1c\x00\x00\x00'),
            'op' / Int32ul,
            'hold_time' / Int32ul,
        )
        # Prebuild required remote messages."""
        msg_open_rc = fmt.build({'op': 1024, 'hold_time': 0})
        msg_key_off = fmt.build({'op': 256, 'hold_time': 0})
        msg_close_rc = fmt.build({'op': 2048, 'hold_time': 0})
        msg_command = fmt.build({'op': op, 'hold_time': hold_time})

        # Send Messages
        self._send_msg(msg_open_rc, encrypted=True)
        time.sleep(0.1)
        self._send_msg(msg_command, encrypted=True)
        self._send_msg(msg_key_off, encrypted=True)
        time.sleep(0.4)
        self._send_msg(msg_close_rc, encrypted=True)
        time.sleep(0.1)

        """recv_msg() is blocking. Is slow returning msg.
        Needs async/thread to receive callback."""
        # msg = self._recv_msg()
        # msg = self._decipher.decrypt(msg)
        # _LOGGER.debug('RX: %s %s', len(msg), binascii.hexlify(msg))
        # return self._handle_response('remote_control', msg)

    def _send_status(self):
        fmt = Struct(
            'length' / Const(b'\x0c\x00\x00\x00'),
            'type' / Const(b'\x14\x00\x00\x00'),
            'status' / Const(b'\x00\x00\x00\x00'),
            'dummy' / Padding(4),
        )

        msg = fmt.build({})
        self._send_msg(msg, encrypted=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Setup for pyps4-homeassistant."""

from setuptools import setup, find_packages
import pyps4_homeassistant.const as pyps4_const

MIN_PY_VERSION = '.'.join(map(str, pyps4_const.REQUIRED_PYTHON_VER))

REQUIRES = [
    'construct>=2.9.45',
    'pycryptodomex>=3.7.2',
]

CLASSIFIERS = [
    'Development Status :: 3 - Alpha',
    'Environment :: Console',
    'License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)',
    'Natural Language :: English',
    'Operating System :: OS Independent',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Topic :: Software Development :: Libraries :: Python Modules',
]

with open('README.rst') as f:
    readme = f.read()

setup(name=pyps4_const.__name__,
      version=pyps4_const.__version__,
      description='PS4 Python Library for Home Assistant',
      long_description=readme,
      author='ktnrg45',
      author_email='ktnrg45@gmail.com',
      packages=find_packages(),
      url='https://github.com/ktnrg45/pyps4-homeassistant',
      license='LGPLv2+',
      classifiers=CLASSIFIERS,
      keywords='playstation sony ps4',
      install_requires=REQUIRES,
      python_requires='>={}'.format(MIN_PY_VERSION),
      test_suite='tests',
      include_package_data=True,
      )



# -*- coding: utf-8 -*-
"""Constants for pyps4-homeassistant."""
MAJOR_VERSION = 0
MINOR_VERSION = 5
PATCH_VERSION = 0
REQUIRED_PYTHON_VER = (3, 5, 3)

__short_version__ = '{}.{}'.format(MAJOR_VERSION, MINOR_VERSION)
__version__ = '{}.{}'.format(__short_version__, PATCH_VERSION)

__name__ = 'pyps4_homeassistant'



# -*- coding: utf-8 -*-
"""Media Art Functions."""
import logging

_LOGGER = logging.getLogger(__name__)

DEPRECATED_REGIONS = {'R1': 'en/us', 'R2': 'en/gb',
                      'R3': 'en/hk', 'R4': 'en/au',
                      'R5': 'en/in'}

# Excluded {China, Japan, Phillipines, Serbia, Ukraine, Vietnam}
COUNTRIES = {"Argentina": "en/ar", "Australia": "en/au", "Austria": "de/at",
             "Bahrain": "en/ae", "Belgium": "fr/be", "Brazil": "en/br",
             "Bulgaria": "en/bg", "Canada": "en/ca", "Chile": "en/cl",
             "Columbia": "en/co", "Costa Rica": "es/cr", "Croatia": "en/hr",
             "Cyprus": "en/cy", "Czech Republic": "en/cz", "Denmark": "en/dk",
             "Ecuador": "es/ec", "El Salvador": "es/sv", "Finland": "en/fi",
             "France": "fr/fr", "Germany": "de/de", "Greece": "en/gr",
             "Guatemala": "es/gt", "Honduras": "es/hn", "Hong Kong": "en/hk",
             "Hungary": "en/hu", "Iceland": "en/is", "India": "en/in",
             "Indonesia": "en/id", "Ireland": "en/ie", "Israel": "en/il",
             "Italy": "it/it", "Korea": "ko/kr", "Kuwait": "en/ae",
             "Lebanon": "en/ae", "Luxembourg": "de/lu", "Maylasia": "en/my",
             "Malta": "en/mt", "Mexico": "en/mx", "Middle East": "en/ae",
             "Nederland": "nl/nl", "New Zealand": "en/nz",
             "Nicaragua": "es/ni", "Norway": "en/no", "Oman": "en/ae",
             "Panama": "es/pa", "Peru": "en/pe", "Poland": "en/pl",
             "Portugal": "pt/pt", "Qatar": "en/ae", "Romania": "en/ro",
             "Russia": "ru/ru", "Saudi Arabia": "en/sa", "Singapore": "en/sg",
             "Slovenia": "en/si", "Slovakia": "en/sk", "South Africa": "en/za",
             "Spain": "es/es", "Sweden": "en/se", "Switzerland": "de/ch",
             "Taiwan": "en/tw", "Thailand": "en/th", "Turkey": "en/tr",
             "United Arab Emirates": "en/ae", "United States": "en/us",
             "United Kingdom": "en/gb"}


def search_all(title, title_id):
    """Search all databases."""
    for x in COUNTRIES:
        region = COUNTRIES[x]
        (_title, art) = get_ps_store_data(title, title_id, region)
        if _title and art:
            return _title, art


def get_ps_store_url(title, region, reformat=False):
    """Get URL for title search in PS Store."""
    import urllib
    import re

    headers = {
        'User-Agent':
            'Mozilla/5.0 '
            '(Windows NT 10.0; Win64; x64) AppleWebKit/537.36 '
            '(KHTML, like Gecko) Chrome/63.0.3239.84 Safari/537.36'
    }

    if title is not None:
        if reformat is True:
            title = re.sub('[^A-Za-z0-9]+', ' ', title)
        title = urllib.parse.quote(title.encode('utf-8'))
        _url = 'https://store.playstation.com/'\
            'valkyrie-api/{0}/19/faceted-search/'\
            '{1}?query={1}&platform=ps4'.format(region, title)

    url = [_url, headers]
    return url


def get_ps_store_data(title, title_id, region, url=None, reformat=False):
    """Get cover art from database."""
    import requests
    import re

    if url is None:
        url = get_ps_store_url(title, region)
    req = None
    match_id = {}
    match_title = {}
    if reformat is True:
        title = re.sub('[^A-Za-z0-9]+', ' ', title)
    type_list = ['Full Game', 'Game', 'PSN Game', 'Bundle', 'App']
    try:
        req = requests.get(url[0], headers=url[1])
        result = req.json()['included']
        if not result:
            title = title.split(' ')
            title = title[0]
            url = get_ps_store_url(title, region)
            req = requests.get(url[0], headers=url[1])
            result = req.json()['included']
    except requests.exceptions.HTTPError as warning:
        _LOGGER.warning("PS cover art HTTP error, %s", warning)
        return
    except requests.exceptions.RequestException as warning:
        _LOGGER.warning("PS cover art request failed, %s", warning)
        return

    # Filter through each item in search request

    # Filter each item by prioritized type
    for game_type in type_list:
        for item in result:
            has_parent = False

            # Set each item as Game object
            game = _game(item)

            # Get Parent attr
            if 'parent' in game:
                if game['parent'] is not None:
                    has_parent = True
                    parent = game['parent']
                    parent_id = _parse_id(parent['id'])
                    parent_title = parent['name']
                    parent_title_f = _format_title(parent_title, reformat)
                    parent_art = "{}/image".format(parent['url'])
                    if parent_id == title_id:
                        _LOGGER.debug("Parent ID Match")
                        return parent_title, parent_art
                    if title.upper() == parent_title_f:
                        _LOGGER.debug("Parent Title Match")
                        return parent_title, parent_art
                    if title.upper() in parent_title_f:
                        _LOGGER.debug("Parent Similar Title Match")
                        return parent_title, parent_art

            if _is_game_type(game, game_type):
                parse_id = _parse_id(game['default-sku-id'])
                title_parse = game['name']

                # If passed SKU matches object SKU
                if parse_id == title_id:
                    cover_art = _get_cover(game)
                    if cover_art is not None:

                        # If true likely a bundle, dlc, deluxe edition
                        if has_parent is False:
                            match_id.update({title_parse: cover_art})

                        # Most likely the intended item so return
                        if title.upper() == _format_title(title_parse,
                                                          reformat):
                            _LOGGER.debug("Direct Match")
                            return title_parse, cover_art

                # Last resort filter if SKU wrong, but title matches.
                elif title.upper() == _format_title(title_parse, reformat):
                    cover_art = _get_cover(game)
                    if cover_art is not None:
                        if has_parent is False:
                            match_title.update({title_parse: cover_art})

    s_title, s_art = _get_similar(title, match_id, match_title, reformat)
    if s_title is None or s_art is None:
        if reformat is False:
            return get_ps_store_data(title, title_id, region, reformat=True)
    return s_title, s_art


def _game(item):
    """Create game object."""
    if 'attributes' in item:
        game = item['attributes']
        return game


def _is_game_type(game, game_type):
    """Check if item is a game and has SKU."""
    if 'game-content-type' in game and \
       game['game-content-type'] == game_type:
        if 'default-sku-id' in game:
            return True


def _parse_id(_id):
    """Parse SKU to simplified ID."""
    try:
        full_id = _id
        full_id = full_id.split("-")
        full_id = full_id[1]
        full_id = full_id.split("_")
        parse_id = full_id[0]
    except IndexError:
        parse_id = "None"
    return parse_id


def _format_title(title, reformat):
    """Format Title."""
    import re

    if reformat is True:
        title = re.sub('[^A-Za-z0-9]+', ' ', title)
    title = title.upper()
    return title


def _get_cover(game):
    """Get cover art."""
    if 'thumbnail-url-base' in game:
        cover = 'thumbnail-url-base'
        cover_art = game[cover]
        return cover_art
    return


def _get_similar(title, match_id, match_title, reformat):
    """Return similar title."""
    if match_id:
        for _title, url in match_id.items():
            if title.upper() in _format_title(_title, reformat):
                cover_art = url
                _LOGGER.debug("Similar Title Match")
                return _title, cover_art
    elif match_title:
        for _title, url in match_title.items():
                cover_art = url
                _LOGGER.debug("Wrong ID Match")
                return _title, cover_art
    return None, None
